/*
 * Copyright (C) 2013
 * 
 * Stefan Fafianie
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * This package makes use of a selection of classes from the libtw library 
 * which is also published under the GNU Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package order;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.List;
import java.util.PriorityQueue;

import libtw.input.GraphInput.InputData;

/* Join node with two children, each from a different part of the tree decomposition but with the same subset.
 * This node can request partial solutions from children and stores new combinations (pairs) of partial solutions in a queue.
 * It returns the partial solution joined from pairs of the smallest weight. 
 * If one of the partial solutions in the pair was the last solution that we received from a child we must request another
 * partial solutions from this child to ensure our ordering.
 */
public class JoinNode<D extends InputData> extends Node<D> {

	Node<D> childLeft, childRight;
	PriorityQueue<Pair> queue;
	boolean askLeft, askRight, cleanupLeft, cleanupRight;
	List<PartialSolution> leftSolutions, rightSolutions;
	
	public JoinNode(BitSet subset, OrderDP<D> dp) {
		super(subset, dp);
		this.subset = subset;
		queue = new PriorityQueue<Pair>();
		askLeft = true;
		askRight = true;
		cleanupLeft = false;
		cleanupRight = false;
		leftSolutions = new ArrayList<PartialSolution>();
		rightSolutions = new ArrayList<PartialSolution>();
	}

	public PartialSolution applyNext() {

		if (askLeft && !cleanupLeft){
			PartialSolution left = childLeft.reduceNext();
			if (left == null){
				cleanupLeft = true;
			}
			else{
				leftSolutions.add(left);
				for (PartialSolution right: rightSolutions){
					queue.add(new Pair(left, right));
				}
			}
		}
		
		if (askRight && !cleanupRight){
			PartialSolution right = childRight.reduceNext();
			if (right == null){
				cleanupRight = true;
			}
			else{
				rightSolutions.add(right);
				for (PartialSolution left: leftSolutions){
					queue.add(new Pair(left, right));
				}
			}
		}
		
		if (queue.isEmpty()){
			hasNext = false;
			return null;
		}
		else{
			Pair p = queue.remove();
			if (leftSolutions.get(leftSolutions.size()-1) == p.left){
				askLeft = true;
				if (!childLeft.hasNext){
					cleanupLeft = true;
				}
			}
			if (rightSolutions.get(rightSolutions.size()-1) == p.right){
				askRight = true;
				if (!childRight.hasNext){
					cleanupRight = true;
				}
			}
			if (cleanupLeft && cleanupRight && queue.isEmpty()){
				hasNext = false;
			}
			dp.operations++;
			return p.left.joinPartialSolution(p.right);
		}
	}

	@Override
	public String toString() {
		return "join"; 
	}

}
