/*
 * Copyright (C) 2013
 * 
 * Stefan Fafianie
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * This package makes use of a selection of classes from the libtw library 
 * which is also published under the GNU Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package order;

/* Alternative to dynamic programming in a bottom-up fashion. This algorithm evaluates partial solutions in order of optimality.
 * Nodes can request a next smallest partial solution from children to calculate the next smallest partial solution for a parent.
 * Once a partial solution is found in the root node we have an optimal solution for Steiner Tree. */
import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import steiner.NTDNiceBag;
import libtw.input.GraphInput.InputData;
import libtw.ngraph.NGraph;
import libtw.ngraph.NVertex;

public class OrderDP<D extends InputData> {

	Node<D> rootNode;
	Set<NVertex<D>> terminals;
	HashMap<Set<NVertex<D>>, Integer> edgeWeights;
	int vertexCount;
	public int operations = 0;
	public int solution;
	
	public OrderDP(){
	}
	
	public void input(int vertexCount, NGraph<NTDNiceBag<D>> g, Set<NVertex<D>> terminals, HashMap<Set<NVertex<D>>, Integer> edgeWeights){
		this.vertexCount = vertexCount;
		NVertex<NTDNiceBag<D>> rootBag = g.getVertex(0);
		BitSet subset = new BitSet(vertexCount);
		traversed = new HashSet<NVertex<NTDNiceBag<D>>>();
		this.terminals = terminals;
		this.edgeWeights = edgeWeights;
		ArrayList<BitSet> subsets = new ArrayList<BitSet>();
		subsets.add(subset);
		rootNode = fixNode(subsets, rootBag).get(0);
	}
	
	Set<NVertex<NTDNiceBag<D>>> traversed;
	
	/* Split bags in the tree decomposition in nodes for each subset of vertices in the bag
	 * Connect nodes that directly depend on each other for partial solutions
	 */
	private List<Node<D>> fixNode(List<BitSet> subsets, NVertex<NTDNiceBag<D>> bag) {
		traversed.add(bag);
		
		ArrayList<NVertex<NTDNiceBag<D>>> children = new ArrayList<NVertex<NTDNiceBag<D>>>();
		for (Iterator<NVertex<NTDNiceBag<D>>> neighbourIterator = bag.getNeighbors(); neighbourIterator.hasNext();){
			NVertex<NTDNiceBag<D>> neighbor = neighbourIterator.next();
			if (!traversed.contains(neighbor)){
				children.add(neighbor);
			}
		}

		List<Node<D>> result = new ArrayList<Node<D>>();
		List<BitSet> newSubsets = new ArrayList<BitSet>();
			
		switch(bag.data.type){
		case LEAF_BAG:
			Node<D> leafNode = new LeafNode<D>(subsets.get(0), this);
			result.add(leafNode);
		
			return result;
		case INTRODUCE_VERTEX_BAG:
			HashMap<BitSet, IntroduceVertexNode<D>> introduceNodes = new HashMap<BitSet, IntroduceVertexNode<D>>();
			HashMap<BitSet, IgnoreVertexNode<D>> ignoreNodes = new HashMap<BitSet, IgnoreVertexNode<D>>();
			
			for (BitSet subset : subsets){
				BitSet dummy = (BitSet) subset.clone();
				if (subset.get(bag.data.vertex.data.id)){
					IntroduceVertexNode<D> intro = new IntroduceVertexNode<D>(subset, bag.data.vertex.data.id, this);
					introduceNodes.put(subset, intro);
					result.add(intro);
					dummy.clear(bag.data.vertex.data.id);
					if (ignoreNodes.containsKey(dummy)){
						IgnoreVertexNode<D> sibling = ignoreNodes.get(dummy);
						intro.siblingIgnore = sibling;
						sibling.siblingIntroduce = intro;
					}
				}
				else{
					IgnoreVertexNode<D> ignore = new IgnoreVertexNode<D>(subset, bag.data.vertex.data.id, terminals.contains(bag.data.vertex), this);
					ignoreNodes.put(subset, ignore);
					newSubsets.add(subset);
					result.add(ignore);
					dummy.set(bag.data.vertex.data.id);
					if (introduceNodes.containsKey(dummy)){
						IntroduceVertexNode<D> sibling = introduceNodes.get(dummy);
						ignore.siblingIntroduce = sibling;
						sibling.siblingIgnore = ignore;						
					}
				}
					
			}
			
			for (Node<D> child: fixNode(newSubsets, children.get(0))){
				IgnoreVertexNode<D> ignore = ignoreNodes.get(child.subset);
				ignore.childNode = child;
			}
			
			return result;
		case FORGET_VERTEX_BAG:
			for (BitSet subset: subsets){
				BitSet dummy = (BitSet) subset.clone();
				dummy.set(bag.data.vertex.data.id);
				ForgetVertexNode<D> forget = new ForgetVertexNode<D>(subset, bag.data.vertex.data.id, this);
				result.add(forget);
				newSubsets.add(subset);
				newSubsets.add(dummy);
			}
			
			List<Node<D>> childNodes = fixNode(newSubsets, children.get(0));
				
			for (int i = 0; i < result.size(); i++){
				((ForgetVertexNode<D>) result.get(i)).childExclude = childNodes.get(i*2);
				((ForgetVertexNode<D>) result.get(i)).childInclude = childNodes.get(i*2+1);
				((ForgetVertexNode<D>) result.get(i)).init();
			}
			
			return result;
		case INTRODUCE_EDGE_BAG:
			for (BitSet subset: subsets){
				Set<NVertex<D>> edge = new HashSet<NVertex<D>>();
				edge.add(bag.data.edge.a);
				edge.add(bag.data.edge.b);
				IntroduceEdgeNode<D> edgeNode = new IntroduceEdgeNode<D>(subset, bag.data.edge.a.data.id, bag.data.edge.b.data.id, edgeWeights.get(edge), this);
				result.add(edgeNode);
				newSubsets.add(subset);
			}
			
			int i = 0;
			
			for (Node<D> child: fixNode(newSubsets, children.get(0))){
				((IntroduceEdgeNode<D>) result.get(i)).childNode = child;
				i++;				
			}
			return result;
		case JOIN_BAG:
			List<BitSet> rightSubsets = new ArrayList<BitSet>();
			for(BitSet subset: subsets){
				BitSet dummy = (BitSet) subset.clone();
				newSubsets.add(subset);
				rightSubsets.add(dummy);
				JoinNode<D> join = new JoinNode<D>(subset, this);
				result.add(join);
			}
			
			List<Node<D>> leftChildNodes = fixNode(newSubsets, children.get(0));
			List<Node<D>> rightChildNodes = fixNode(rightSubsets, children.get(1));
			
			i = 0;
			
			for (Node<D> join: result){
				((JoinNode<D>) join).childLeft = leftChildNodes.get(i);
				((JoinNode<D>) join).childRight = rightChildNodes.get(i);
				i++;
			}
			
			return result;
		default:
			throw new Error("unknown bag type");	
		}
	}
	
	public void run(){
		//Let the rootnode ask request the smallest partial solution (i.e. the optimal solution for Steiner Tree)
		solution = ((ForgetVertexNode<D>) rootNode).childInclude.reduceNext().cost;
	}

	
	
}
