/*
 * Copyright (C) 2013
 * 
 * Stefan Fafianie
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * This package makes use of a selection of classes from the libtw library 
 * which is also published under the GNU Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package steiner;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.Set;

import order.OrderDP;
import libtw.algorithm.Constructive;
import libtw.algorithm.GreedyDegree;
import libtw.algorithm.PermutationToTreeDecomposition;
import libtw.input.InputException;
import libtw.input.GraphInput.InputData;
import libtw.ngraph.NGraph;
import libtw.ngraph.NTDBag;
import libtw.ngraph.NVertex;

/* Helper class for batched experiments. Give a folder path as input to perform 
 * experiments on all .stp instances contained inside. Results are gathered for
 * each algorithm. We measure time and the amount of generated partial solutions. */
public class BatchExperiments {

	public static void main(String[] args) throws IOException, InputException {
		
		StpReader in;
		NGraph<InputData> g;
		FileWriter writer = new FileWriter("results.csv");
		writer.append("instance;tw;|V|;|E|;|T|;FL_size;FL_time;RBA_size;RBA_time;RBC_size;RBC_time;ROWA_time;ROWA_size;ROWC_time;ROWC_size;SPLIT_time;SPLIT_size");
		writer.append("\n");
		
		File dir = new File(args[0]);
		for (File file: dir.listFiles()){
			String name = file.getName();
			System.out.println("processing " + name);
			
			in = new StpReader( dir.getAbsolutePath() +"/" + name);
			g = in.get();
			
			Constructive<InputData> theAlgorithm = new PermutationToTreeDecomposition<InputData>( new GreedyDegree<InputData>() );
			theAlgorithm.setInput(g);
			theAlgorithm.run();
			NGraph<NTDBag<InputData>> td = theAlgorithm.getDecomposition();	
			
			int tw = 0;
			for (int i = 0; i < td.getNumberOfVertices(); i++){
				if (td.getVertex(i).data.vertices.size()>tw){
					tw = td.getVertex(i).data.vertices.size();
				}
			}
			tw--;
			
			int v = g.getNumberOfVertices();
			int e = g.getNumberOfEdges();
			
			Set<NVertex<InputData>> terminals = in.getTerminals();
			int t = terminals.size();
			NVertex<InputData> terminal = terminals.iterator().next();
			HashMap<Set<NVertex<InputData>>, Integer> weights = in.getWeights();
			
			TreeDecompositionToNiceTreeDecomposition<InputData> ntd = new TreeDecompositionToNiceTreeDecomposition<InputData>(terminal);
			ntd.setInput(td);
			ntd.run();		
			NGraph<NTDNiceBag<InputData>> niceDecomposition = ntd.getNiceDecomposition();
			
			SteinerTreeDP<InputData> dp = new SteinerTreeDP<InputData>();
			dp.setInput(niceDecomposition, terminals, weights);
			
			long time = System.nanoTime();
			dp.run();
			long FL_time = System.nanoTime()-time;
			int FL_size = dp.totalsize;
			
			System.out.println("folklore - finished");
			
			RankBasedSteinerTreeDP<InputData> rbdp = new RankBasedSteinerTreeDP<InputData>();
			rbdp.setInput(niceDecomposition, terminals, weights);
			rbdp.setCondition(true);
			
			time = System.nanoTime();
			rbdp.run();
			long RB1_time = System.nanoTime()-time;
			int RB1_size = rbdp.totalsize;
			
			System.out.println("rank based always - finished");
			
			rbdp = new RankBasedSteinerTreeDP<InputData>();
			rbdp.setInput(niceDecomposition, terminals, weights);
			rbdp.setCondition(false);
			
			time = System.nanoTime();
			rbdp.run();
			long RB2_time = System.nanoTime()-time;
			int RB2_size = rbdp.totalsize;
			
			System.out.println("rank based conditional - finished");
			
			RowDP<InputData> row = new RowDP<InputData>();
			row.setInput(niceDecomposition, terminals, weights);
			row.always = true;
			
			time = System.nanoTime();
			row.run();
			long ROWA_time = System.nanoTime()-time;
			long ROWA_size = row.operations;
			
			System.out.println("row always - finished");
			
			row = new RowDP<InputData>();
			row.setInput(niceDecomposition, terminals, weights);
			row.always = false;
					
			time = System.nanoTime();
			row.run();
			long ROWC_time = System.nanoTime()-time;
			long ROWC_size = row.operations;
			
			System.out.println("row conditional - finished");
			
			OrderDP<InputData> split = new OrderDP<InputData>();
			
			time = System.nanoTime();
			split.input(g.getNumberOfVertices(), niceDecomposition, terminals, weights);
			split.run();
			long SPLIT_time = System.nanoTime()-time;
			int SPLIT_size = split.operations;
			
			System.out.println("split - finished");
			
			writer.append(name + ";" + 
						  tw + ";" +
						  v + ";" +
						  e + ";" + 
						  t + ";" +
						  FL_size + ";" +
						  FL_time + ";" +
						  RB1_size + ";" +
						  RB1_time + ";" +
						  RB2_size + ";" +
						  RB2_time + ";" +
						  ROWA_time + ";" +
						  ROWA_size + ";" +
						  ROWC_time + ";" +
						  ROWC_size + ";" +
						  SPLIT_time + ";" +
						  SPLIT_size + ";" 
						  );
			writer.append("\n");
						  
			System.out.println(name + ";" + 
					  tw + ";" +
					  v + ";" +
					  e + ";" + 
					  t + ";" +
					  FL_size + ";" +
					  FL_time + ";" +
					  RB1_size + ";" +
					  RB1_time + ";" +
					  RB2_size + ";" +
					  RB2_time + ";" +
					  ROWA_time + ";" +
					  ROWA_size + ";" +
					  ROWC_time + ";" +
					  ROWC_size + ";" +
					  SPLIT_time + ";" +
					  SPLIT_size + ";" 
					  );
		}
		
		writer.flush();
		writer.close();
	}
}
