/*
 * Copyright (C) 2013
 * 
 * Stefan Fafianie
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * This package makes use of a selection of classes from the libtw library 
 * which is also published under the GNU Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package steiner;

import java.util.*;

import libtw.input.GraphInput.InputData;
import libtw.ngraph.NVertex;
import libtw.ngraph.NEdge;

/* Connectivity representation as a nested set (partition). The basic operators 
 * that are required in the DP formulation are provided here. */
public class Connectivity<D extends InputData> implements Comparable<Connectivity<D>>{
	
	Set<Set<NVertex<D>>> partition;
	int cost;
	
	//Create an empty connectivity.
	public Connectivity(){
		partition = new HashSet<Set<NVertex<D>>>();
	}
	
	//Add a vertex as singleton in this connectivity.
	public Connectivity<D> addVertex(NVertex<D> vertex){
		Connectivity<D> result = new Connectivity<D>();
		result.partition.addAll(partition);
		Set<NVertex<D>> block = new HashSet<NVertex<D>>();
		block.add(vertex);
		result.partition.add(block);	
		return result;
	}
	
	/* Forget a vertex in this connectivity. Returns null if the vertex 
	 * was singleton, unless it is the only vertex left in the bag 
	 * (allows for easy extraction of the final solution). */
	public Connectivity<D> forgetVertex(NVertex<D> vertex, boolean usedAllVertices){
		Connectivity<D> result = new Connectivity<D>();
		boolean valid = false;
		for(Set<NVertex<D>> block: partition){
			HashSet<NVertex<D>> newBlock = new HashSet<NVertex<D>>();
			newBlock.addAll(block);
			if (newBlock.contains(vertex)){
				if(newBlock.size() > 1 || partition.size() == 1 && usedAllVertices){
					newBlock.remove(vertex);
					if (newBlock.size() != 0){
						result.partition.add(newBlock);
					}
					valid = true;
				}
			}
			else{
				result.partition.add(newBlock);
			}
		}
		return valid ? result: null;
	}
	
	/* Add an edge in this connectivity. Blocks containing 
	 * both incident vertices are joined together. */
	public Connectivity<D> addEdge(NEdge<D> edge){
		Connectivity<D> result = new Connectivity<D>();	
		Set<NVertex<D>> blocka = null, blockb = null;
		for(Set<NVertex<D>> block: partition){		
			if (block.contains(edge.a)){
				blocka = block;
			}
			if(block.contains(edge.b)){
				blockb = block;
			}
			if(!block.contains(edge.a) && !block.contains(edge.b)){
				result.partition.add(block);
			}	
		}				
		Set<NVertex<D>> newBlock = new HashSet<NVertex<D>>();
		newBlock.addAll(blocka);
		newBlock.addAll(blockb);
		result.partition.add(newBlock);	
		return result;
	}
	
	/* Join this connectivity with another. Breath-first 
	 * search is used to do this in linear time. */
	public Connectivity<D> joinConnectivity(Connectivity<D> otherConnectivity){
		Set<Set<NVertex<D>>> resultPartition = new HashSet<Set<NVertex<D>>>();
		
		HashMap<NVertex<D>, Boolean> checklist = new HashMap<NVertex<D>, Boolean>();
		HashMap<NVertex<D>, Set<NVertex<D>>> leftBlockMap = new HashMap<NVertex<D>,Set<NVertex<D>>>();
		HashMap<NVertex<D>, Set<NVertex<D>>> rightBlockMap = new HashMap<NVertex<D>,Set<NVertex<D>>>();
		
		for(Set<NVertex<D>> block:this.partition){
			for(NVertex<D> vertex:block){
				leftBlockMap.put(vertex, block);
				checklist.put(vertex, false);
			}
		}		
		for(Set<NVertex<D>> block:otherConnectivity.partition){
			for(NVertex<D> vertex:block){
				rightBlockMap.put(vertex, block);
				checklist.put(vertex,  false);
			}
		}
		
		for (NVertex<D> vertex: checklist.keySet()){
			if (!checklist.get(vertex)){
				Set<NVertex<D>> newBlock = new HashSet<NVertex<D>>();
				newBlock.add(vertex);		
				Boolean depleted = false;
				while(!depleted){
					depleted = true;
					Set<NVertex<D>> newVertices = new HashSet<NVertex<D>>();
					for(NVertex<D> v: newBlock){
						if (!checklist.get(v)){
							depleted = false;
							checklist.put(v, true);
							if (leftBlockMap.containsKey(v)){
								newVertices.addAll(leftBlockMap.get(v));
							}
							if (rightBlockMap.containsKey(v)){
								newVertices.addAll(rightBlockMap.get(v));
							}
						}
					}
					newBlock.addAll(newVertices);
				}
				resultPartition.add(newBlock);					
			}
		}			
		
		Connectivity<D> result = new Connectivity<D>();
		result.partition = resultPartition;
				
		return result;
	}
	
	/* Test if this connectivity is a refinement of another. 
	 * Returns false at the first counterexample. */
	public boolean isRefinement(Connectivity<D> otherConnectivity){
		boolean refined = true;
		
		Iterator<Set<NVertex<D>>> thisPartition = partition.iterator();
		while (thisPartition.hasNext() && refined){
			Set<NVertex<D>> block = thisPartition.next();
			boolean present = false;
			Iterator<Set<NVertex<D>>> otherPartition = otherConnectivity.partition.iterator();
			while (otherPartition.hasNext() && !present){
				Set<NVertex<D>> block2 = otherPartition.next();
				if (block2.containsAll(block)){
					present = true;
				}
			}
			if (!present){
				refined = false;
			}
		}
		
		return refined;
	}
	
	/* Add hash-codes for vertices in blocks and multiply hash-codes for blocks.
	 * This counters the commutativity of the standard approach for sets in java,
	 * which in the nested case adds everything. Uses a big prime to make sure
	 * we don't get an integer overflow.
	 */
	@Override
	public int hashCode(){
		int code = 0;	
		for(Set<NVertex<D>> vertexset: partition){			
			int blockcode = 1;
			for(NVertex<D> vertex: vertexset){
				blockcode = blockcode*(vertex.hashCode()%2147483647);
			}
			code += blockcode;
		}			
		return code;
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Object obj){
		if (this == obj)
	        return true;
	    if (obj == null)
	        return false;
	    if (getClass() != obj.getClass()){
	        return false;
	    }
	    else{
			final Connectivity<D> otherConnectivity = (Connectivity<D>) obj;
	    	if (partition.equals(otherConnectivity.partition))
	    		return true;
	    }
	    return false;
	}
	
	//For debugging.
	public void printConnectivity(){
		String output = "{ ";
		for(Set<NVertex<D>> vertexset: partition){
			output += "{ ";
			for(NVertex<D> vertex: vertexset){
				output+= vertex.data.name + " ";
			}
			output += "} ";		
		}
		System.out.println(output + "}" );
	}

	//For sorting.
	@Override
	public int compareTo(Connectivity<D> other) {
		return this.cost - other.cost;
	}

	
}
