/*
 * Copyright (C) 2013
 * 
 * Stefan Fafianie
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * This package makes use of a selection of classes from the libtw library 
 * which is also published under the GNU Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package steiner;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import libtw.input.GraphInput.InputData;
import libtw.ngraph.NEdge;
import libtw.ngraph.NGraph;
import libtw.ngraph.NVertex;

/* Implementation of the dynamic programming algorithm for Steiner-tree using weighted bit strings to represent partial solutions.
 * This is mostly a direct adaptation of the SteinerTreeDP class with small changes to account for the new representation. 
 * See the difference between Connectivity used in SteinerTreeDP and Row used here. Partial solutions are now directly represented
 * by rows in the cutmatrices. This allows us to calculate cutmatrices of parent nodes from cutmatrices of child nodes.*/

public class RowDP<D extends InputData> {

	private NGraph<NTDNiceBag<D>> niceDecomposition;
	private Set<NVertex<D>> terminals;
	private HashMap<Set<NVertex<D>>, Integer> edgeWeights;
	boolean reduce = true;
	boolean always = false;
	
	public long operations = 0;
	public long solution = 0;
	
	public void setInput(NGraph<NTDNiceBag<D>> g, Set<NVertex<D>> terminals, HashMap<Set<NVertex<D>>, Integer> edgeWeights) {
		this.niceDecomposition = g;	
		this.terminals = terminals;
		this.edgeWeights = edgeWeights;
	}
	
	public void run(){
		traversed = new HashSet<NVertex<NTDNiceBag<D>>>();
		NVertex<NTDNiceBag<D>> root = niceDecomposition.getVertex(0);
		traversed.add(root);
		HashMap<Set<NVertex<D>>,HashMap<Row,Integer>> bag = calculateBag(root.getNeighbors().next());
		
		//Extract the solution from the single partial solution contained in the child node table
		solution = bag.get(bag.keySet().iterator().next()).entrySet().iterator().next().getValue();
	}
	
	Set<NVertex<NTDNiceBag<D>>> traversed;	
	public HashMap<Set<NVertex<D>>,HashMap<Row,Integer>> calculateBag(NVertex<NTDNiceBag<D>> bag){	
		traversed.add(bag);
		
		// Fetch the children for this bag.
		ArrayList<NVertex<NTDNiceBag<D>>> children = new ArrayList<NVertex<NTDNiceBag<D>>>();
		for (Iterator<NVertex<NTDNiceBag<D>>> neighborIterator = bag.getNeighbors(); neighborIterator.hasNext();){
			NVertex<NTDNiceBag<D>> neighbor = neighborIterator.next();
			if (!traversed.contains(neighbor)){
				children.add(neighbor);
			}
		}
		
		HashMap<Set<NVertex<D>>,HashMap<Row,Integer>> result;
		
		// Calculate the partial solution table for this bag.
		switch(bag.data.type){
		case LEAF_BAG:
			result = leafBag();
			break;
		case INTRODUCE_VERTEX_BAG:
			result = introduceVertexBag(bag.data.vertex, terminals.contains(bag.data.vertex), calculateBag(children.get(0)));
			break;
		case FORGET_VERTEX_BAG:
			result = forgetVertexBag(bag.data.vertex, calculateBag(children.get(0)));
			break;
		case INTRODUCE_EDGE_BAG:
			result = introduceEdgeBag(bag.data.edge, calculateBag(children.get(0)));
			break;
		default:
			result = joinBag(calculateBag(children.get(0)), calculateBag(children.get(1)));		
			break;
		}
		
		//Apply the rank based reductions
		if (reduce){
			reduce(result);
		}
				
		return result;
	}
	
	private void reduce(
			HashMap<Set<NVertex<D>>, HashMap<Row, Integer>> solutionTable) {

		for (Set<NVertex<D>> used:solutionTable.keySet()){
			
			boolean apply = always ? (solutionTable.get(used).size() > 1) 
					: ( solutionTable.get(used).size() > Math.max(1, Math.pow(2, used.size()-1)));
			
			if (apply){

				ArrayList<Row> rows = new ArrayList<Row>();
				rows.addAll(solutionTable.get(used).keySet());
				Collections.sort(rows);
				HashMap<Row, Integer> subTable = new HashMap<Row, Integer>();
					
				//We can now fill cutmatrices by directly copying values from the partial solutions.
				//This is where we save time
				int[][] matrix = new int[rows.size()][rows.get(0).size];
				for (int i = 0; i < rows.size(); i++){
					for (int j = 0; j < rows.get(0).size; j++){
						if (rows.get(i).row.get(j)) {
							matrix[i][j] = 1;
							
						}
						else{
							matrix[i][j] = 0;
							
						}
					}
				}
								
				//Keep track of columns that are to be processed.
				ArrayList<Integer> swipePositions = new ArrayList<Integer>();
				for (int i = 0; i < rows.get(0).size; i++){
					swipePositions.add(i);
				}
				
				int i = 0;
				//We can stop when we have found a partial solution for each column.
				while (i < rows.size() && swipePositions.size() > 0){
			
					//Find the leading 1.
					boolean hasOne = false;
					int swipePosition = -1;
					int j = 0;
					while(j < rows.get(0).size && !hasOne){
						if (matrix[i][j] == 1){
							hasOne = true;
							swipePosition = j;
						}
						j++;
					
					}
					
					/* Disregard the partial solution if it's row has all 0's 
					 * (linearly dependent on solutions that where found earlier). */
					if (hasOne){
						//Add the partial solution to the result table.
						
						subTable.put(rows.get(i), rows.get(i).cost);
						
						
						//Swipe the column of it's leading 1.
						for (int k = i+1; k < rows.size(); k++){
							if (matrix[k][swipePosition] == 1){
								//Only swipe columns that did not contain a leading 1 previously.
								for (j = 0; j < swipePositions.size(); j++){
									int column = swipePositions.get(j);
									int oldValue = matrix[k][column];
									int swipeValue = matrix[i][column];
									int newValue = (oldValue + swipeValue) % 2;
									matrix[k][column] = newValue;
									
								}								
							}
						}		
					}
					
					//Kick out the column with leading 1.
					swipePositions.remove((Integer)swipePosition);
					i++;
				}
							
				solutionTable.put(used, subTable);
				
			}
		}
	}

	private HashMap<Set<NVertex<D>>, HashMap<Row, Integer>> introduceEdgeBag(
			NEdge<D> introducedEdge,
			HashMap<Set<NVertex<D>>, HashMap<Row, Integer>> childTable) {
		
		HashMap<Set<NVertex<D>>,HashMap<Row,Integer>> solutionTable = new HashMap<Set<NVertex<D>>,HashMap<Row,Integer>>();
		
		//don't use the edge
		for (Set<NVertex<D>> used: childTable.keySet()){
			HashMap<Row, Integer> childRowTable = childTable.get(used);
			HashMap<Row, Integer> rowTable = new HashMap<Row, Integer>();
			for (Row row: childRowTable.keySet()){
				rowTable.put(row, row.cost);
			}
			solutionTable.put(used, rowTable);			
		}		
		
		//use the edge if it improves anything
		for (Set<NVertex<D>> used: childTable.keySet()){
			if (used.contains(introducedEdge.a) && used.contains(introducedEdge.b)){
				HashMap<Row, Integer> childRowTable = childTable.get(used);
				HashMap<Row, Integer> rowTable = solutionTable.get(used);
				Set<NVertex<D>> edge = new HashSet<NVertex<D>>();
				edge.add(introducedEdge.a);
				edge.add(introducedEdge.b);
				
				Row edgePrint = Row.edgePrint(used.size(), findIndex(introducedEdge.a, used), findIndex(introducedEdge.b, used), edgeWeights.get(edge));
				for (Row row: childRowTable.keySet()){
					this.operations++;
					Row newRow = row.joinRow(edgePrint);
					if (rowTable.containsKey(newRow)){
						
						newRow.cost = Math.min(newRow.cost, rowTable.get(newRow));
						rowTable.remove(newRow);
					}
					rowTable.put(newRow, newRow.cost);
				}
			}
		}
		
		return solutionTable;
	}

	/* Used to find the index of a particular vertex in a set of vertices ordered by their ID.
	 * We need this for the new operators, since we make strict assumptions about the position
	 * of cuts based on this information. */
	private int findIndex(NVertex<D> v, Set<NVertex<D>> used) {
		int result = 0;
		for (NVertex<D> u: used){
			if (u.data.id < v.data.id){
				result++;
			}
		}
		return result;
	}

	private HashMap<Set<NVertex<D>>, HashMap<Row, Integer>> joinBag(
			HashMap<Set<NVertex<D>>, HashMap<Row, Integer>> leftChildTable,
			HashMap<Set<NVertex<D>>, HashMap<Row, Integer>> rightChildTable) {

		HashMap<Set<NVertex<D>>,HashMap<Row,Integer>> solutionTable = new HashMap<Set<NVertex<D>>,HashMap<Row,Integer>>();
		
		Set<Set<NVertex<D>>> jointKeySet = new HashSet<Set<NVertex<D>>>();
		jointKeySet.addAll(leftChildTable.keySet());
		jointKeySet.addAll(rightChildTable.keySet());
		
		for (Set<NVertex<D>> used: jointKeySet){
			HashMap<Row, Integer> leftChildRowTable = leftChildTable.get(used);
			HashMap<Row, Integer> rightChildRowTable = rightChildTable.get(used);
			HashMap<Row, Integer> rowTable = new HashMap<Row, Integer>();
			
			for (Row leftRow: leftChildRowTable.keySet()){
				for (Row rightRow: rightChildRowTable.keySet()){
					Row newRow = leftRow.joinRow(rightRow);
					this.operations++;
					if (rowTable.containsKey(newRow)){
						newRow.cost = Math.min(newRow.cost,  rowTable.get(newRow));
						rowTable.remove(newRow);
					}
					rowTable.put(newRow, newRow.cost);
					
				}
			}
			solutionTable.put(used, rowTable);
		}
		
		return solutionTable;
	}

	private HashMap<Set<NVertex<D>>, HashMap<Row, Integer>> forgetVertexBag(
			NVertex<D> forgottenVertex,
			HashMap<Set<NVertex<D>>, HashMap<Row, Integer>> childTable) {
		
		HashMap<Set<NVertex<D>>,HashMap<Row,Integer>> solutionTable = new HashMap<Set<NVertex<D>>,HashMap<Row,Integer>>();
		
		List<Set<NVertex<D>>> usedForgottenVertex = new ArrayList<Set<NVertex<D>>>();
		List<Set<NVertex<D>>> unusedForgottenVertex = new ArrayList<Set<NVertex<D>>>();
		
		for (Set<NVertex<D>> used: childTable.keySet()){
			if (used.contains(forgottenVertex)){
				usedForgottenVertex.add(used);
			}
			else{
				unusedForgottenVertex.add(used);
			}
			
		}
		
		// commit to using v.
		for (Set<NVertex<D>> used: usedForgottenVertex){
			HashMap<Row, Integer> childRowTable = childTable.get(used);
			HashMap<Row, Integer> rowTable = new HashMap<Row, Integer>();
			Set<NVertex<D>> newUsed = new HashSet<NVertex<D>>();

			for (Row childRow: childRowTable.keySet()){
				Row newRow = childRow.forgetVertex(used.size(), findIndex(forgottenVertex, used));
				if (newRow != null){
					this.operations++;
					if (rowTable.containsKey(newRow)){
						newRow.cost = Math.min(newRow.cost, rowTable.get(newRow));
						rowTable.remove(newRow);
					}
					rowTable.put(newRow, newRow.cost);
		
				}
			}
			
			newUsed.addAll(used);
			newUsed.remove(forgottenVertex);
			
			solutionTable.put(newUsed, rowTable);

			
		}
		
		// commit to not using v.
		for (Set<NVertex<D>> used: unusedForgottenVertex){
			HashMap<Row, Integer> childRowTable = childTable.get(used);
			if (solutionTable.containsKey(used)){
				HashMap<Row, Integer> previousRowTable = solutionTable.get(used);
				for (Row childRow:childRowTable.keySet()){
					if (previousRowTable.containsKey(childRow)){
						childRow.cost = Math.min(childRow.cost, previousRowTable.get(childRow));
						previousRowTable.remove(childRow);
					}
					previousRowTable.put(childRow, childRow.cost);
				}
				
			}
			else{
				solutionTable.put(used,  childRowTable);
			}
		}
		return solutionTable;
	}

	private HashMap<Set<NVertex<D>>, HashMap<Row, Integer>> introduceVertexBag(
			NVertex<D> introducedVertex, boolean isTerminal,
			HashMap<Set<NVertex<D>>, HashMap<Row, Integer>> childTable) {
		
		HashMap<Set<NVertex<D>>,HashMap<Row,Integer>> solutionTable = new HashMap<Set<NVertex<D>>,HashMap<Row,Integer>>();
		
		// use v
		for (Set<NVertex<D>> used: childTable.keySet()){
			Set<NVertex<D>> newUsed = new HashSet<NVertex<D>>();
			newUsed.addAll(used);
			newUsed.add(introducedVertex);
			HashMap<Row, Integer> rowTable = new HashMap<Row, Integer>();
			
				HashMap<Row, Integer> childRowTable = childTable.get(used);
				for (Row childRow: childRowTable.keySet()){
					int cost = childRowTable.get(childRow);
					this.operations++;
					Row newRow = childRow.addVertex(findIndex(introducedVertex, used), newUsed.size());
					rowTable.put(newRow,  cost);
				
			}
			
			solutionTable.put(newUsed, rowTable);
		}
		
		// don't use v
	    if(!isTerminal){
	    	for(Set<NVertex<D>> used: childTable.keySet()){
	    		solutionTable.put(used, childTable.get(used));
	    	}
	    }
		
		return solutionTable;
	}

	HashMap<Set<NVertex<D>>,HashMap<Row,Integer>> leafBag(){
		HashMap<Set<NVertex<D>>,HashMap<Row,Integer>> solutionTable = new HashMap<Set<NVertex<D>>,HashMap<Row,Integer>>();
		HashMap<Row, Integer> rowTable = new HashMap<Row, Integer>();
		Set<NVertex<D>> used = new HashSet<NVertex<D>>();
		this.operations++;
		rowTable.put(Row.leafRow(), 0);
		solutionTable.put(used, rowTable);
		return solutionTable;
	}

}
