/*
 * Copyright (C) 2013
 * 
 * Stefan Fafianie
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * This package makes use of a selection of classes from the libtw library 
 * which is also published under the GNU Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package steiner;

import java.io.*;
import java.util.*;

import libtw.input.GraphInput;
import libtw.input.InputException;
import libtw.ngraph.NGraph;
import libtw.ngraph.NVertex;
import libtw.ngraph.ListGraph;
import libtw.ngraph.ListVertex;

/* A quick and dirty line-reader used to read .stp files. */
public class StpReader implements GraphInput {

	protected String filename;
	protected NGraph<InputData> g;
	protected Set<NVertex<InputData>> terminals;
	protected HashMap<Set<NVertex<InputData>>, Integer> edgeWeights;
	protected HashMap<String, NVertex<InputData>> vertices;
	
	public StpReader(String filename){
		this.filename = filename;
	}
		
	public Set<NVertex<InputData>> getTerminals(){
		return terminals;
	}
	
	public HashMap<Set<NVertex<InputData>>, Integer> getWeights(){
		return edgeWeights;
	}
	
	@Override
	public NGraph<InputData> get() throws InputException {
		
		g = new ListGraph<InputData>();
		
		FileReader fr;
		try	{
			fr = new FileReader(filename);
		}
		catch (FileNotFoundException e)	{
			throw new InputException(e);
		}
		LineNumberReader in = new LineNumberReader( fr );
		String line;
		try{
			line = in.readLine();

			String[] tokens = line.split("\\s+");
			if (!tokens[0].toLowerCase().equals("33d32945") ){
				throw new Error("not an stp file");
			}		
		}
		catch (IOException e) {
			throw new InputException(e);
		}
		
		Boolean reading = true;
		
		try{
			while( (line=in.readLine()) !=null && reading){
				
				String[] tokens = line.split("\\s+");
				assert tokens.length > 0;
				
				String command = tokens[0];
				
				if( command.equals("") )
				{	// empty line
					continue;
				} else if (command.toLowerCase().equals("section")) {
					String section = tokens[1].toLowerCase();
					if (section.equals("comment")){
						skipSection(in);
					} else if (section.equals("graph")){
						readGraph(in);
					} else if (section.equals("terminals")){
						readTerminals(in);
					} else if (section.equals("coordinates")){
						skipSection(in);
					} else{
						throw new Error("unknown section");
					}
				} else if (command.equals("EOF")) {
					reading = false;
				} else {
					System.out.println( "Unknown command '" + command + "' at line " + in.getLineNumber() );
				}
			}
		}
		catch (IOException e) {
			throw new InputException(e);
		}		
		return g;
	}
	
	private void skipSection(LineNumberReader in) throws IOException{	
		String line;
		Boolean reading = true;
		while( (line=in.readLine()) !=null && reading){	
			String[] tokens = line.split("\\s+");
			String command = tokens[0].toLowerCase();	
			if (command.equals("end")){
				reading = false;
			} 
		}
	}
	
	private void readGraph(LineNumberReader in) throws IOException{
		vertices = new HashMap<String, NVertex<InputData>>();
		edgeWeights = new HashMap<Set<NVertex<InputData>>, Integer>();
		
		String line;
		Boolean reading = true;
		int currentEdges = 0;
		int edges = 0;
		while( (line=in.readLine()) !=null && reading){	
			String[] tokens = line.split("\\s+");
			String command = tokens[0].toLowerCase();	
			if (command.equals("nodes")){
				for (int i = 0; i < Integer.parseInt(tokens[1]); i++){
					NVertex<InputData> newVertex = new ListVertex<InputData>(new InputData(vertices.size(),""+(i+1)));
					vertices.put(""+(i+1), newVertex);
					g.addVertex(newVertex);
				}
			} else if (command.equals("edges")) {
				edges = Integer.parseInt(tokens[1]);
			} else if (command.equals("e")){
				currentEdges++;
				NVertex<InputData> v1 = vertices.get(tokens[1]);
				NVertex<InputData> v2 = vertices.get(tokens[2]);
				int weight = Integer.parseInt(tokens[3]);
				g.addEdge(v1, v2);
				Set<NVertex<InputData>> edge = new HashSet<NVertex<InputData>>();
				edge.add(v1);
				edge.add(v2);
				edgeWeights.put(edge, weight);
			} else if (command.equals("arcs")) {
				throw new Error("arcs are currently not supported");
			} else if (command.equals("a")){
				throw new Error("arcs are currently not supported");
			}
			else if (command.equals("end")){
				reading = false;
			} else{
				System.out.println( "Unknown command '" + command + "' at line " + in.getLineNumber() );
			}
		}
		
		if (edges != currentEdges){
			throw new Error("missing edges");
		}
	}
	
	private void readTerminals(LineNumberReader in) throws IOException{
		terminals = new HashSet<NVertex<InputData>>();
		String line;
		Boolean reading = true;
		int currentTerminals = 0;
		int totalTerminals = 0;
		while( (line=in.readLine()) !=null && reading){	
			String[] tokens = line.split("\\s+");
			String command = tokens[0].toLowerCase();	
			if (command.equals("terminals")){
				totalTerminals = Integer.parseInt(tokens[1]);
			} else if (command.equals("t")){
				terminals.add(vertices.get(tokens[1]));
				currentTerminals++;
			} else if (command.equals("end")){
				reading = false;
			} else{
				System.out.println( "Unknown command '" + command + "' at line " + in.getLineNumber() );
			}
		}
		if (totalTerminals != currentTerminals){
			throw new Error("missing terminals");
		}
	}
	

}
