/*
 * Copyright (C) 2013
 * 
 * Stefan Fafianie
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * This package makes use of a selection of classes from the libtw library 
 * which is also published under the GNU Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package steiner;

import java.util.*;

import libtw.input.GraphInput.InputData;
import libtw.ngraph.*;

import steiner.NTDNiceBag.NiceBagType;

// Use this class to transform a tree decomposition into a nice tree decomposition.
public class TreeDecompositionToNiceTreeDecomposition<D extends InputData>{

	NGraph<NTDBag<D>> decomposition;
	NGraph<NTDNiceBag<D>> niceDecomposition;
	NVertex<D> forgetRoot;
	
	public void setInput(NGraph<NTDBag<D>> g) {
		this.decomposition = g;	
	}

	public void run() {
		niceDecomposition = new ListGraph<NTDNiceBag<D>>();
		
		NVertex<NTDBag<D>> rootBag = null;
		
		// Find a root bag.
		if (forgetRoot != null){
			Iterator<NVertex<NTDBag<D>>> decompositionBagIterator = decomposition.getVertices();
			while (decompositionBagIterator.hasNext() && rootBag == null ){
				NVertex<NTDBag<D>> bag = decompositionBagIterator.next();
				if (bag.data.vertices.contains(forgetRoot)){
					rootBag = bag;
				}
			}
		}
		else{
			rootBag = decomposition.getVertex(0);
		}
		
		forgottenVertices = new HashSet<NVertex<D>>();
		traversed = new HashSet<NVertex<NTDBag<D>>>();
		
		// Fix the root.
		ArrayList<NVertex<D>> bottomBagContents = new ArrayList<NVertex<D>>();
		bottomBagContents.addAll(rootBag.data.vertices);
		Set<NVertex<D>> topBagContents = new HashSet<NVertex<D>>();
		NVertex<NTDNiceBag<D>> previousNiceBagVertex = unfoldForgetBags(null, topBagContents, bottomBagContents);
		
		// Traverse from the root bag.		
		fixBag(rootBag, previousNiceBagVertex);
	}
	
	//Keep track of bags that have been processed (tree decomposition is represented in a graph, so we need to make sure not to fix parents again).
	Set<NVertex<NTDBag<D>>> traversed;
	
	private void fixBag(NVertex<NTDBag<D>> decompositionBag, NVertex<NTDNiceBag<D>> previousNiceDecompositionBag) {
		traversed.add(decompositionBag);
		
		// Fetch the children for this bag.
		ArrayList<NVertex<NTDBag<D>>> children = new ArrayList<NVertex<NTDBag<D>>>();
		for (Iterator<NVertex<NTDBag<D>>> neighbourIterator = decompositionBag.getNeighbors(); neighbourIterator.hasNext();){
			NVertex<NTDBag<D>> neighbor = neighbourIterator.next();
			if (!traversed.contains(neighbor)){
				children.add(neighbor);
			}
		}
		
		switch (children.size()){
		
		case 0:
			// Fix leaf node by adding a series of introduce vertex bags.
			ArrayList<NVertex<D>> topBagContents = new ArrayList<NVertex<D>>();
			topBagContents.addAll(decompositionBag.data.vertices);
			Set<NVertex<D>> bottomBagContents = new HashSet<NVertex<D>>();
			
			NVertex<NTDNiceBag<D>> previousNiceBagVertex = unfoldIntroduceBags(previousNiceDecompositionBag, bottomBagContents, topBagContents);
					
			// Add the nice leaf bag at the end.
			NTDNiceBag<D> leafBag = new NTDNiceBag<D>();
			leafBag.type = NiceBagType.LEAF_BAG;
			NVertex<NTDNiceBag<D>> leafBagVertex = new ListVertex<NTDNiceBag<D>>();
			leafBagVertex.data = leafBag;
			niceDecomposition.addVertex(leafBagVertex);
			niceDecomposition.addEdge(leafBagVertex,  previousNiceBagVertex);		
			break;
			
		case 1:
			// Fix this single child node with forget/introduce nodes.	
			Set<NVertex<D>> intersection = new HashSet<NVertex<D>>();
			intersection.addAll(decompositionBag.data.vertices);
			intersection.retainAll(children.get(0).data.vertices);
			
			Set<NVertex<D>> parentOnly = new HashSet<NVertex<D>>();
			parentOnly.addAll(decompositionBag.data.vertices);
			parentOnly.removeAll(children.get(0).data.vertices);
				
			Set<NVertex<D>> childOnly = new HashSet<NVertex<D>>();
			childOnly.addAll(children.get(0).data.vertices);
			childOnly.removeAll(decompositionBag.data.vertices);
				
			// Add a series of introduce vertex bags until we have the intersection.
			ArrayList<NVertex<D>> parentOnlyArray = new ArrayList<NVertex<D>>();
			parentOnlyArray.addAll(parentOnly);
			previousNiceBagVertex = unfoldIntroduceBags(previousNiceDecompositionBag, intersection, parentOnlyArray);
			
			// Add a series of forget vertex bags until we have the child node.
			ArrayList<NVertex<D>> childOnlyBag = new ArrayList<NVertex<D>>();
			childOnlyBag.addAll(childOnly);			
			previousNiceBagVertex = unfoldForgetBags(previousNiceBagVertex, intersection, childOnlyBag);
			
			// Fix the child node
			fixBag(children.get(0), previousNiceBagVertex);
			break;
			
		default:
			// Fix join node.
			NVertex<NTDBag<D>> oldLeftChild = children.remove(0);
			decompositionBag.removeNeighbor(oldLeftChild);
			for (NVertex<NTDBag<D>> child: children){
				decompositionBag.removeNeighbor(child);
			}
			
			ArrayList<NVertex<D>> bagContents = new ArrayList<NVertex<D>>();
			bagContents.addAll(decompositionBag.data.vertices);
			
			NVertex<NTDBag<D>> leftChild = new ListVertex<NTDBag<D>>();
			NVertex<NTDBag<D>> rightChild = new ListVertex<NTDBag<D>>();
			leftChild.data = new NTDBag<D>();
			rightChild.data = new NTDBag<D>();
			
			leftChild.data.vertices.addAll(bagContents);
			rightChild.data.vertices.addAll(bagContents);
			
			// Split the node in the tree decomposition.
			decomposition.addVertex(leftChild);
			decomposition.addVertex(rightChild);
			decomposition.addEdge(decompositionBag, leftChild);
			decomposition.addEdge(decompositionBag, rightChild);
			decomposition.addEdge(leftChild, oldLeftChild);
			for (NVertex<NTDBag<D>> child: children){
				decomposition.addEdge(rightChild, child);
			}
			
			// Add the join bag node.
			NTDNiceBag<D> niceBag = new NTDNiceBag<D>();
			niceBag.type = NiceBagType.JOIN_BAG;
			niceBag.vertices.addAll(bagContents);
			
			// Add the new bag to the nice decomposition.
			ListVertex<NTDNiceBag<D>> niceBagVertex = new ListVertex<NTDNiceBag<D>>();
			niceBagVertex.data = niceBag;
			niceDecomposition.addVertex(niceBagVertex);
			
			// Add edges between bags.
			niceDecomposition.addEdge(niceBagVertex, previousNiceDecompositionBag);
			
			// Fix left child.
			fixBag(leftChild, niceBagVertex);
			
			// Fix right child.
			fixBag(rightChild, niceBagVertex);
			break;
		}
		
	}
	
	// Helper function for unfolding forget vertex bags.
	HashSet<NVertex<D>> forgottenVertices;
	public NVertex<NTDNiceBag<D>> unfoldForgetBags(NVertex<NTDNiceBag<D>> previousNiceDecompositionBag, Set<NVertex<D>> topBagContents, ArrayList<NVertex<D>> bottomBagContents){
		
		NVertex<D> forgottenVertex;
		NVertex<NTDNiceBag<D>> niceBagVertex;
		NVertex<NTDNiceBag<D>> previousNiceBagVertex = previousNiceDecompositionBag;
		
		Set<NVertex<D>> helperBag = new HashSet<NVertex<D>>();
		
		// Add a forget vertex bag for every vertex in the root bag.
		while (bottomBagContents.size() > 0){
			
			// Begin with the preselected vertex if specified, iterate over bag.
			if (forgetRoot != null && niceDecomposition.getNumberOfVertices()==0){
				forgottenVertex = forgetRoot;
			}
			else{
				forgottenVertex = bottomBagContents.get(0);
			}
			
			// Set up the new bag.
			bottomBagContents.remove(forgottenVertex);
			NTDNiceBag<D> niceBag = new NTDNiceBag<D>();
			niceBag.type = NiceBagType.FORGET_VERTEX_BAG;
			niceBag.vertex = forgottenVertex;
			niceBag.vertices.addAll(helperBag);
			niceBag.vertices.addAll(topBagContents);
			helperBag.add(forgottenVertex);		
					
			// Add the new bag to the nice decomposition.
			niceBagVertex = new ListVertex<NTDNiceBag<D>>();
			niceBagVertex.data = niceBag;
			niceDecomposition.addVertex(niceBagVertex);
			
			// Add edges between bags.
			if (previousNiceBagVertex != null){
				niceDecomposition.addEdge(niceBagVertex, previousNiceBagVertex);
			}		
			
			forgottenVertices.add(forgottenVertex);
			previousNiceBagVertex = niceBagVertex;
			
			
			// Add introduce edge bags for pairs of vertices that have been forgotten before.
			for(Iterator<NVertex<D>> neighbourIterator = forgottenVertex.getNeighbors(); neighbourIterator.hasNext();){
				NVertex<D> neighbor = neighbourIterator.next();
				if (forgottenVertices.contains(neighbor)){
					
					// Set up an introduce edge bag.
					niceBag = new NTDNiceBag<D>();
					niceBag.type = NiceBagType.INTRODUCE_EDGE_BAG;
					niceBag.edge = new NEdge<D>(forgottenVertex, neighbor);
					niceBag.vertices.addAll(helperBag);
					niceBag.vertices.addAll(topBagContents);
					
					// Add the new bag to the nice decomposition.
					niceBagVertex = new ListVertex<NTDNiceBag<D>>();
					niceBagVertex.data = niceBag;
					niceDecomposition.addVertex(niceBagVertex);
					niceDecomposition.addEdge(niceBagVertex, previousNiceBagVertex);
					
					previousNiceBagVertex = niceBagVertex;
				}
			}
			
		}
		
		return previousNiceBagVertex;
	}
	
	// Helper function for unfolding introduce bags.
	public NVertex<NTDNiceBag<D>> unfoldIntroduceBags(NVertex<NTDNiceBag<D>> previousNiceDecompositionBag, Set<NVertex<D>> bottomBagContents, ArrayList<NVertex<D>> topBagContents){
		
		NVertex<D> introducedVertex;
		NVertex<NTDNiceBag<D>> niceBagVertex;
		NVertex<NTDNiceBag<D>> previousNiceBagVertex = previousNiceDecompositionBag;
		
		// Add an introduce vertex bag for every vertex in this leaf bag.
		while (topBagContents.size() > 0){
			introducedVertex = topBagContents.get(0);
			
			// Set up the new bag.
			NTDNiceBag<D> niceBag = new NTDNiceBag<D>();
			niceBag.type = NiceBagType.INTRODUCE_VERTEX_BAG;
			niceBag.vertex = introducedVertex;
			niceBag.vertices.addAll(topBagContents);
			niceBag.vertices.addAll(bottomBagContents);
			topBagContents.remove(introducedVertex);
			
			// Add the new bag to the nice decomposition.
			niceBagVertex = new ListVertex<NTDNiceBag<D>>();
			niceBagVertex.data = niceBag;
			niceDecomposition.addVertex(niceBagVertex);
			
			// Add edges between bags.	
			niceDecomposition.addEdge(niceBagVertex, previousNiceBagVertex);
			
			previousNiceBagVertex = niceBagVertex;
		}
		
		return previousNiceBagVertex;
	}	

	//Set a specific vertex to get a nice decomposition rooted at the forget bag for this vertex.
	public TreeDecompositionToNiceTreeDecomposition(NVertex<D> forgetRoot)
	{
		this.forgetRoot = forgetRoot;
	}
	
	public NGraph<NTDNiceBag<D>> getNiceDecomposition()
	{
		return niceDecomposition;
	}

}
